import java.util.Arrays;
 
/**
 * A vectorial implementation of a queue.
 * 
 * @author Luis S. Reis
 * @version $Id: ArrayQueue.java 1767 2013-06-11 15:58:21Z asantos $
 * @param <E> parameter of the elements of the queue
 */
public final class ArrayQueue<E> implements Cloneable {

    /**
     *  The initial capacity of the vectorial queue
     */
    private static final int DEFAULT_CAPACITY = 3;

    /**
     *  The elements in this queue
     */
    private E[] theQueue;

   /**
    * Head of the queue, i.e. index of the first element to exit the queue 
    */
    private int front;

    /**
     * Rear of the queue, i.e. index of the last element entered into the queue
     */
    private int rear;

    /**
     *  The number of elements in this queue
     */
    private int size;

    /**
     * Constructor for an empty Queue
     */
    public ArrayQueue() {
    	clear();
    }
    
	/**
	 * Remove all elements from this queue.
	 */
	@SuppressWarnings("unchecked")
	public void clear() {
		theQueue = (E[]) new Object[DEFAULT_CAPACITY];
		front = 0;
		rear = theQueue.length - 1;
		size = 0;
	}
	
    /**
     * Inserts an item at the end of the Queue
     * 
     * @param item - item being inserted
     */
    public void offer(E item) {
        if (size == theQueue.length) {
            grow();
        }
        rear = (rear + 1) % theQueue.length;
        theQueue[rear] = item;
        size++;
    }

    /**
     * The first item in the queue
     * 
     * @return the first item
     */
    public E element() {
        return theQueue[front];
    }

    /**
     * Removes the first item in the Queue
     */
    public void remove() {
        theQueue[front] = null;
        front = (front + 1) % theQueue.length;
        size--;
    }
	
	/**
	 * Tests if the queue is empty
	 * 
	 * @return true if the queue is empty
	 */
    public boolean isEmpty() {
        return size == 0;
    }

    /**
     * Returns a clone of this Queue
     * 
     * @return clone of the queue
     */
    @SuppressWarnings("unchecked")
	public ArrayQueue<E> clone() {
        try {
            ArrayQueue<E> result = (ArrayQueue<E>) super.clone();
            result.theQueue = theQueue.clone();
            return result;
        } catch (CloneNotSupportedException e) {
            throw new InternalError(e.getMessage());
        }
    }

	/**
     * The String representation of the queue
     * 
     * @return String representation of the queue
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("");
        for (int i = 0; i < theQueue.length; i++)
        	sb.append(theQueue[i] + " ");
        return sb.toString();
    }
    
	/**
     * Hashcode for the queue
     * 
     * @return hashcode for the queue
     */
    @Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + front;
		result = prime * result + rear;
		result = prime * result + size;
		result = prime * result + Arrays.hashCode(theQueue);
		return result;
	}

    /**
     * Is this Queue equal to a given object?
     * 
     * @param other object to be compared
     * @return true if the Queue is equal to the other object, false otherwise
     */
    @SuppressWarnings("unchecked")
    @Override
	public boolean equals(Object other) {
        return this == other || (other != null &&
                other instanceof ArrayQueue<?>
                && ((ArrayQueue<E>) other).size == size
                && equalElements(((ArrayQueue<E>) other).theQueue,
                        ((ArrayQueue<E>) other).front));
    }
    
    /**
     * Are the elements in this queue equal to those in a given array?
     * 
     * @param other The array to compare
     * @param j - front of the queue being compared
     * @return true if the elements are equal
     * @requires other.length >= size
     */
    private boolean equalElements(E[] other, int j) {
        int i = front;
        for (int k = 0; k < size; k++) {
            if (theQueue[i] == null || other[j] == null || !theQueue[i].equals(other[j])) {
                return false;
            }
            i = (i + 1) % theQueue.length;
            j = (j + 1) % other.length;
        }
        return true;
    }

	/**
     * Increase the capacity of the array used to store the elements.
     */
    @SuppressWarnings("unchecked")
	private void grow() {
		E[] newArray= (E[]) new Object[theQueue.length * 2];
        for (int j = 0; j < size; j++) {
            newArray[j] = theQueue[front];
            front = (front + 1) % theQueue.length;
        }
        front = 0;
        rear = size - 1;
        theQueue = newArray;
    }
}
